import json

import jsonschema

import shell_whiz.openai_client as openai
from shell_whiz.exceptions import (
    EditingError,
    ExplanationError,
    TranslationError,
    WarningError,
)
from shell_whiz.llm_jsonschemas import (
    dangerous_command_jsonschema,
    edited_shell_command_jsonschema,
    translation_jsonschema,
)


async def suggest_shell_command(prompt, preferences, model):
    translation = await openai.suggest_shell_command(
        prompt=prompt, preferences=preferences, model=model
    )

    try:
        translation_json = json.loads(translation)
    except json.JSONDecodeError:
        raise TranslationError("Could not extract JSON.")

    try:
        jsonschema.validate(
            instance=translation_json, schema=translation_jsonschema
        )
    except jsonschema.ValidationError:
        raise TranslationError("Generated JSON is not valid.")

    shell_command = translation_json["shell_command"].strip()

    if shell_command == "":
        raise TranslationError("Extracted shell command is empty.")

    return shell_command


async def recognize_dangerous_command(shell_command, preferences, model):
    dangerous_command = await openai.recognize_dangerous_command(
        shell_command=shell_command, preferences=preferences, model=model
    )

    try:
        dangerous_command_json = json.loads(dangerous_command)
    except json.JSONDecodeError:
        raise WarningError("Could not extract JSON.")

    try:
        jsonschema.validate(
            instance=dangerous_command_json,
            schema=dangerous_command_jsonschema,
        )
    except jsonschema.ValidationError:
        raise WarningError("Generated JSON is not valid.")

    is_dangerous = dangerous_command_json["dangerous_to_run"]
    dangerous_consequences = dangerous_command_json.get(
        "dangerous_consequences", ""
    ).strip()

    if not is_dangerous:
        return False, ""
    elif dangerous_consequences == "":
        raise WarningError("Extracted dangerous consequences are empty.")
    elif "\n" in dangerous_consequences:
        raise WarningError("Extracted dangerous consequences contain newlines.")
    else:
        return True, dangerous_consequences


async def get_explanation_of_shell_command(
    shell_command, preferences, model, stream=False
):
    return await openai.get_explanation_of_shell_command(
        shell_command=shell_command,
        preferences=preferences,
        model=model,
        stream=stream,
    )


async def get_explanation_of_shell_command_by_chunks(
    shell_command=None, preferences=None, model=None, stream=None
):
    is_first_chunk = True
    skip_initial_spaces = True
    async for chunk in openai.get_explanation_of_shell_command_by_chunks(
        shell_command=shell_command,
        preferences=preferences,
        model=model,
        stream=stream,
    ):
        if chunk is None:
            break

        if skip_initial_spaces:
            chunk = chunk.lstrip()
            if chunk:
                skip_initial_spaces = False
            else:
                continue

        if is_first_chunk:
            if not chunk.startswith("*"):
                raise ExplanationError("Explanation is not valid.")
            is_first_chunk = False

        yield chunk


async def edit_shell_command(shell_command, prompt, model):
    edited_shell_command = await openai.edit_shell_command(
        shell_command=shell_command, prompt=prompt, model=model
    )

    try:
        edited_shell_command_json = json.loads(edited_shell_command)
    except json.JSONDecodeError:
        raise EditingError("Could not extract JSON.")

    try:
        jsonschema.validate(
            instance=edited_shell_command_json,
            schema=edited_shell_command_jsonschema,
        )
    except jsonschema.ValidationError:
        raise EditingError("Generated JSON is not valid.")

    edited_shell_command = edited_shell_command_json[
        "edited_shell_command"
    ].strip()
    if edited_shell_command == "":
        raise EditingError("Edited shell command is empty.")

    return edited_shell_command



import json
import os

import questionary
import rich

from shell_whiz.constants import ERROR_PREFIX_RICH


def get_config_paths():
    config_dir = os.path.join(
        (
            os.environ.get("XDG_CONFIG_HOME")
            or os.environ.get("APPDATA")
            or os.path.join(os.environ.get("HOME", os.getcwd()), ".config")
        ),
        "shell-whiz",
    )
    config_file = os.path.join(config_dir, "config.json")

    return config_dir, config_file


async def edit_config_cli():
    rich.print(
        "Visit https://platform.openai.com/account/api-keys to get your API key."
    )
    openai_api_key = await questionary.text(
        "OpenAI API key",
        default=os.environ.get("OPENAI_API_KEY", ""),
        validate=lambda text: len(text) > 0,
    ).unsafe_ask_async()

    return {"OPENAI_API_KEY": openai_api_key}


async def edit_config():
    config_dir, config_file = get_config_paths()
    config = await edit_config_cli()

    try:
        os.makedirs(config_dir)
        json.dump(config, config_file)
    except OSError:
        rich.print(f"{ERROR_PREFIX_RICH}: Couldn't write to file {config_file}")

    try:
        os.chmod(config_file, 0o600)
    except OSError:
        rich.print(
            f"{ERROR_PREFIX_RICH}: Failed to change file permissions for {config_file}"
        )

    return config


def read_config():
    _, config_file = get_config_paths()

    try:
        with open(config_file) as f:
            config = json.load(f)
    except (OSError, json.JSONDecodeError):
        return {}
    else:
        return config


async def configure():
    config = read_config()
    if "OPENAI_API_KEY" not in config:
        config = await edit_config()

    os.environ["OPENAI_API_KEY"] = config["OPENAI_API_KEY"]



from openai import AsyncOpenAI

from shell_whiz.llm_jsonschemas import (
    dangerous_command_jsonschema,
    edited_shell_command_jsonschema,
    translation_jsonschema,
)


def format_preferences(preferences):
    return f"These are my preferences: ####\n{preferences}\n####"


async def suggest_shell_command(prompt, preferences, model):
    client = AsyncOpenAI()
    response = await client.chat.completions.create(
        model=model,
        temperature=0.25,
        max_tokens=256,
        messages=[
            {
                "role": "user",
                "content": f"{format_preferences(preferences)}\n\n{prompt}",
            }
        ],
        functions=[
            {
                "name": "perform_task_in_command_line",
                "description": "Perform a task in the command line",
                "parameters": translation_jsonschema,
            }
        ],
        function_call={"name": "perform_task_in_command_line"},
    )
    return response.choices[0].message.function_call.arguments


async def recognize_dangerous_command(shell_command, preferences, model):
    client = AsyncOpenAI()
    response = await client.chat.completions.create(
        model=model,
        temperature=0,
        max_tokens=96,
        messages=[
            {
                "role": "user",
                "content": f"{format_preferences(preferences)}\n\nI want to run this command: ####\n{shell_command}\n####",
            },
        ],
        functions=[
            {
                "name": "recognize_dangerous_command",
                "description": "Recognize a dangerous shell command. This function should be very low sensitive, only mark a command dangerous when it has very serious consequences.",
                "parameters": dangerous_command_jsonschema,
            }
        ],
        function_call={"name": "recognize_dangerous_command"},
    )
    return response.choices[0].message.function_call.arguments


async def get_explanation_of_shell_command(
    shell_command, preferences, model, stream
):
    prompt = f'Split the command into parts and explain it in **list** format. Each line should follow the format "command part" followed by an explanation.\n\nFor example, if the command is `ls -l`, you would explain it as:\n* `ls` lists directory contents.\n  * `-l` displays in long format.\n\nFor `cat file | grep "foo"`, the explanation would be:\n* `cat file` reads the content of `file`.\n* `| grep "foo"` filters lines containing "foo".\n\n* Never explain basic command line concepts like pipes, variables, etc.\n* Keep explanations clear, simple, concise and elegant (under 7 words per line).\n* Use two spaces to indent for each nesting level in your list.\n\nIf you can\'t provide an explanation for a specific shell command or it\'s not a shell command, you should reply with an empty JSON object.\n\n{format_preferences(preferences)}\n\nShell command: ####\n{shell_command}\n####'

    temperature = 0.1
    max_tokens = 512

    client = AsyncOpenAI()
    return await client.chat.completions.create(
        model=model,
        temperature=temperature,
        max_tokens=max_tokens,
        stream=stream,
        messages=[{"role": "user", "content": prompt}],
    )


async def get_explanation_of_shell_command_by_chunks(
    shell_command=None, preferences=None, model=None, stream=None
):
    if stream is None:
        stream = await get_explanation_of_shell_command(
            shell_command=shell_command,
            preferences=preferences,
            model=model,
            stream=True,
        )

    async for chunk in stream:
        yield chunk.choices[0].delta.content


async def edit_shell_command(shell_command, prompt, model):
    client = AsyncOpenAI()
    response = await client.chat.completions.create(
        model=model,
        temperature=0.2,
        max_tokens=256,
        messages=[
            {
                "role": "user",
                "content": f"{shell_command}\n\nPrompt: ####\n{prompt}\n####",
            },
        ],
        functions=[
            {
                "name": "edit_shell_command",
                "description": "Edit a shell command, according to the prompt",
                "parameters": edited_shell_command_jsonschema,
            }
        ],
        function_call={"name": "edit_shell_command"},
    )
    return response.choices[0].message.function_call.arguments



translation_jsonschema = {
    "type": "object",
    "properties": {
        "shell_command": {
            "type": "string",
            "description": "Shell command to perform the task, leave blank if it is not possible to perform the task in the command line",
        }
    },
    "required": ["shell_command"],
}

dangerous_command_jsonschema = {
    "type": "object",
    "properties": {
        "dangerous_to_run": {"type": "boolean"},
        "dangerous_consequences": {
            "type": "string",
            "description": "Brief explanation of the potential side effects of running the command. No more than 12 words.",
        },
    },
    "required": ["dangerous_to_run"],
}

edited_shell_command_jsonschema = {
    "type": "object",
    "properties": {
        "edited_shell_command": {
            "type": "string",
            "description": "Edited shell command, leave blank if you couldn't edit it",
        }
    },
    "required": ["edited_shell_command"],
}



import asyncio
import subprocess
import sys

import openai
import questionary
import rich
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown

from shell_whiz.argparse import create_argument_parser
from shell_whiz.config import configure, edit_config
from shell_whiz.constants import ERROR_PREFIX_RICH, THINKING_MSG
from shell_whiz.exceptions import (
    EditingError,
    ExplanationError,
    TranslationError,
    WarningError,
)
from shell_whiz.llm_client import (
    edit_shell_command,
    get_explanation_of_shell_command,
    get_explanation_of_shell_command_by_chunks,
    recognize_dangerous_command,
    suggest_shell_command,
)

console = Console()


def print_command(shell_command):
    rich.print(
        "\n ==================== [bold green]Command[/] ====================\n"
    )
    print(" " + " ".join(shell_command.splitlines(keepends=True)) + "\n")


async def check_danger(shell_command, preferences, model):
    with console.status(
        "Shell Whiz is checking the command for danger...", spinner="dots"
    ):
        try:
            return await recognize_dangerous_command(
                shell_command, preferences, model
            )
        except WarningError:
            return False, ""


async def print_explanation(
    shell_command=None,
    preferences=None,
    explain_using=None,
    stream=None,
    insert_newline=False,
):
    if insert_newline:
        print()

    rich.print(
        " ================== [bold green]Explanation[/] =================="
    )

    try:
        with Live("", auto_refresh=False) as live:
            explanation = ""
            async for chunk in get_explanation_of_shell_command_by_chunks(
                shell_command=shell_command,
                preferences=preferences,
                model=explain_using,
                stream=stream,
            ):
                explanation += chunk
                live.update(Markdown(explanation), refresh=True)
    except ExplanationError:
        rich.print(
            f" {ERROR_PREFIX_RICH}: Sorry, I don't know how to explain this command."
        )

    print()


async def edit_shell_command_cli(shell_command, prompt, model):
    try:
        with console.status(THINKING_MSG, spinner="dots"):
            shell_command = await edit_shell_command(
                shell_command, prompt, model
            )
    except EditingError:
        rich.print(
            f"\n {ERROR_PREFIX_RICH}: Sorry, I couldn't edit the command. I left it unchanged."
        )

    return shell_command


def get_filtered_choices(dont_explain, explain_using):
    choices = [
        "Run this command",
        "Explain this command",
        "Explain using GPT 3.5 Turbo [1106]",
        "Explain using GPT-4 Turbo [BETA]",
        "Explain using GPT-4",
        "Revise query",
        "Edit manually",
        "Exit",
    ]

    if not dont_explain:
        choices.remove("Explain this command")

    if explain_using == "gpt-3.5-turbo-1106":
        choices.remove("Explain using GPT 3.5 Turbo [1106]")
    elif explain_using == "gpt-4-1106-preview":
        choices.remove("Explain using GPT-4 Turbo [BETA]")
    elif explain_using == "gpt-4":
        choices.remove("Explain using GPT-4")

    return choices


async def perform_selected_action(
    shell_command,
    is_dangerous,
    preferences,
    explain_using,
    shell,
    dont_explain,
    output_file,
):
    while True:
        choice = await questionary.select(
            "Select an action",
            get_filtered_choices(dont_explain, explain_using),
        ).unsafe_ask_async()

        if choice == "Exit":
            sys.exit(1)
        elif choice == "Run this command":
            if output_file:
                try:
                    with open(output_file, "w", newline="\n") as f:
                        f.write(shell_command)
                except OSError:
                    sys.exit(1)
            else:
                cancel_run = (
                    is_dangerous
                    and not await questionary.confirm(
                        "Are you sure you want to run this command?"
                    ).unsafe_ask_async()
                )
                if cancel_run:
                    sys.exit(1)
                subprocess.run(shell_command, executable=shell, shell=True)
            # End successfully only if the command has been executed
            sys.exit()
        elif choice == "Explain this command":
            await print_explanation(
                shell_command=shell_command,
                preferences=preferences,
                explain_using=explain_using,
                insert_newline=True,
            )
        elif choice == "Explain using GPT 3.5 Turbo [1106]":
            await print_explanation(
                shell_command=shell_command,
                preferences=preferences,
                explain_using="gpt-3.5-turbo-1106",
                insert_newline=True,
            )
        elif choice == "Explain using GPT-4 Turbo [BETA]":
            await print_explanation(
                shell_command=shell_command,
                preferences=preferences,
                explain_using="gpt-4-1106-preview",
                insert_newline=True,
            )
        elif choice == "Explain using GPT-4":
            await print_explanation(
                shell_command=shell_command,
                preferences=preferences,
                explain_using="gpt-4",
                insert_newline=True,
            )
        elif choice == "Revise query":
            edit_prompt = (
                await questionary.text(
                    message="Enter your revision"
                ).unsafe_ask_async()
            ).strip()
            if edit_prompt != "":
                return shell_command, edit_prompt
        elif choice == "Edit manually":
            edited_shell_command = (
                await questionary.text(
                    "Edit command",
                    default=shell_command,
                    multiline="\n" in shell_command,
                ).unsafe_ask_async()
            ).strip()
            if (
                edited_shell_command != ""
                and edited_shell_command != shell_command
            ):
                return edited_shell_command, ""


async def run_ai_assistant(
    prompt,
    preferences,
    model,
    explain_using,
    dont_explain,
    dont_warn,
    quiet,
    shell,
    output_file,
):
    await configure()

    try:
        with console.status(THINKING_MSG, spinner="dots"):
            shell_command = await suggest_shell_command(
                prompt=prompt, preferences=preferences, model=model
            )
    except TranslationError:
        rich.print(f"{ERROR_PREFIX_RICH}: Sorry, I don't know how to do this.")
        sys.exit(1)

    edit_prompt = ""
    while True:
        if edit_prompt != "":
            shell_command = await edit_shell_command_cli(
                shell_command=shell_command, prompt=edit_prompt, model=model
            )

        if not dont_explain:
            print_command(shell_command)
            stream_task = asyncio.create_task(
                get_explanation_of_shell_command(
                    shell_command=shell_command,
                    preferences=preferences,
                    model=explain_using,
                    stream=True,
                )
            )

        if dont_warn:
            is_dangerous = False
        else:
            (
                is_dangerous,
                dangerous_consequences,
            ) = await check_danger(
                shell_command=shell_command,
                preferences=preferences,
                model=model,
            )

        if dont_explain:
            print_command(shell_command)

        if not dont_warn and is_dangerous:
            rich.print(
                " [bold red]Warning[/]: [bold yellow]{0}[/]\n".format(
                    dangerous_consequences
                )
            )

        if not dont_explain:
            await print_explanation(stream=await stream_task)

        if quiet:
            break

        shell_command, edit_prompt = await perform_selected_action(
            shell_command=shell_command,
            is_dangerous=is_dangerous,
            preferences=preferences,
            explain_using=explain_using,
            shell=shell,
            dont_explain=dont_explain,
            output_file=output_file,
        )


async def parse_arguments():
    args = create_argument_parser().parse_args()

    if not sys.stdin.isatty() or not sys.stdout.isatty():
        try:
            rich.print(
                f"{ERROR_PREFIX_RICH}: Shell Whiz cannot run in non-interactive mode."
            )
        except BrokenPipeError:
            pass

        sys.exit(1)

    if args.command == "config":
        await edit_config()
    elif args.command == "ask":
        prompt = " ".join(args.prompt).strip()
        if prompt == "":
            rich.print(f"{ERROR_PREFIX_RICH}: Please provide a valid prompt.")
            sys.exit(1)

        await run_ai_assistant(
            prompt=prompt,
            preferences=args.preferences,
            model=args.model,
            explain_using=args.explain_using or args.model,
            dont_explain=args.dont_explain,
            dont_warn=args.dont_warn,
            quiet=args.quiet,
            shell=args.shell,
            output_file=args.output,
        )


def run():
    try:
        asyncio.run(parse_arguments())
    except KeyboardInterrupt:
        print("\nExiting...")
        sys.exit(1)
    except openai.BadRequestError:
        rich.print(
            f"{ERROR_PREFIX_RICH}: Your request was malformed or missing some required parameters, such as a token or an input."
        )
        sys.exit(1)
    except openai.AuthenticationError:
        rich.print(
            f"{ERROR_PREFIX_RICH}: You are not authorized to access the OpenAI API. You may have entered the wrong API key. Your API key is invalid, expired or revoked. Please run [bold green]sw config[/] to set up the API key. Visit https://platform.openai.com/account/api-keys to get your API key."
        )
        sys.exit(1)
    except openai.PermissionDeniedError:
        rich.print(
            f"{ERROR_PREFIX_RICH}: Your API key or token does not have the required scope or role to perform the requested action. Make sure your API key has the appropriate permissions for the action or model accessed."
        )
        sys.exit(1)
    except openai.RateLimitError:
        rich.print(
            f"{ERROR_PREFIX_RICH}: OpenAI API request exceeded rate limit. If you are on a free plan, please upgrade to a paid plan for a better experience with Shell Whiz. Visit https://platform.openai.com/account/billing/limits for more information."
        )
        sys.exit(1)
    except openai.APITimeoutError:
        rich.print(
            f"{ERROR_PREFIX_RICH}: OpenAI API request timed out. Please retry your request after a brief wait."
        )
        sys.exit(1)
    except openai.APIConnectionError:
        rich.print(
            f"{ERROR_PREFIX_RICH}: OpenAI API request failed to connect. Please check your internet connection and try again."
        )
        sys.exit(1)
    except openai.InternalServerError:
        rich.print(
            f"{ERROR_PREFIX_RICH}: OpenAI API request failed due to a temporary server-side issue. Please retry your request after a brief wait. The problem is on the side of the OpenAI. Visit https://status.openai.com for more information."
        )
        sys.exit(1)
    except openai.APIStatusError:
        rich.print(
            f"{ERROR_PREFIX_RICH}: An error occurred while connecting to the OpenAI API. Please retry your request after a brief wait. The problem is on the side of the OpenAI. Visit https://status.openai.com for more information."
        )
        sys.exit(1)
    except openai.APIError:
        rich.print(
            f"{ERROR_PREFIX_RICH}: An unknown error occurred while connecting to the OpenAI API. Please retry your request after a brief wait."
        )
        sys.exit(1)



class ShellWhizError(Exception):
    pass


class TranslationError(ShellWhizError):
    pass


class WarningError(ShellWhizError):
    pass


class ExplanationError(ShellWhizError):
    pass


class EditingError(ShellWhizError):
    pass



import argparse
from importlib.metadata import version

from shell_whiz.constants import SW_DESCRIPTION


def create_argument_parser():
    parser = argparse.ArgumentParser(description=SW_DESCRIPTION)

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"{SW_DESCRIPTION}. Version {version('shell-whiz')}.",
    )

    subparsers = parser.add_subparsers(dest="command", required=True)

    subparsers.add_parser("config", help="Configure Shell Whiz")

    ask_parser = subparsers.add_parser(
        "ask", help="Ask Shell Whiz for a shell command"
    )
    ask_parser.add_argument(
        "-s", "--shell", type=str, help="set the shell executable"
    )
    ask_parser.add_argument(
        "-p",
        "--preferences",
        type=str,
        default="I use Bash on Linux",
        help="set your preferences (default: %(default)s)",
    )
    ask_parser.add_argument(
        "-m",
        "--model",
        type=str,
        default="gpt-3.5-turbo",
        help="select the model to use",
    )
    ask_parser.add_argument(
        "--explain-using", type=str, help="select the model to explain"
    )
    ask_parser.add_argument(
        "-n",
        "--dont-explain",
        action="store_true",
        help="don't explain the command",
    )
    ask_parser.add_argument(
        "--dont-warn",
        action="store_true",
        help="don't warn about dangerous commands",
    )
    ask_parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="don't show the menu, end immediately",
    )
    ask_parser.add_argument(
        "-o", "--output", type=str, help="output file to write the command to"
    )
    ask_parser.add_argument(
        "prompt", nargs="+", type=str, help="Prompt for Shell Whiz"
    )

    return parser
